
from collections import OrderedDict
from collections import namedtuple


class flux_row_cls:

    class_names = {'_headers',
                   'is_bound',
                   'values'}

    def __init__(self, headers, values):
        """
        :param headers: OrderedDict of {'header': index}
        :param values:  list of underlying data

        headers is a single dictionary passed byref from the
        flux_cls to all flux_row_cls instances, reducing unneccesary
        memory usage and allowing all mapping updates to be made
        instantaneously

        namedtuples may be more efficient, but too much of a headache to deal
        with their immutability

        self.__dict__ is used to set attributes in __init__ so as to avoid
        premature __setattr__ lookups
        """
        self.__dict__['_headers'] = headers
        self.__dict__['values']   = values
        self.__dict__['is_bound'] = False

    @property
    def names(self):
        return list(self._headers.keys())

    @property
    def view_as_array(self):
        """ for development purposes; used purely to trigger a debugging feature in PyCharm

        PyCharm will recognize returned the (name, value) pairs as an ndarray
        and enable the "...view as array" option in the debugger which displays
        values in a special window
        """
        import numpy
        return numpy.transpose([self.names, self.values])

    def dict(self):
        return OrderedDict(zip(self.names, self.values))

    def namedtuples(self):
        try:
            nt_cls = namedtuple('flux_row_nt', self.names)
            return nt_cls(*self.values)
        except ValueError as e:
            import re

            names = [n for n in self.names
                       if re.search('^[^a-z]|[ ]', n, re.IGNORECASE)]
            raise ValueError("invalid headers for namedtuple: {}".format(names)) from e

    def bind(self):
        """ bind headers / values directly to instance, avoiding need for subsequent __getattr__ lookups
        (only use this if you know the side-effects)
        """
        if self.__dict__['is_bound']:
            return

        self.__dict__.update(zip(self.names, self.values))

        self.__dict__['values']   = '(values bound to __dict__)'
        self.__dict__['is_bound'] = True

    def unbind(self, names=None):
        if not self.__dict__['is_bound']:
            return

        if names is None:
            names = self.__dict__.keys() - self.class_names

        values = self.__dict__['values'] = []
        for k in names:
            v = self.__dict__.pop(k)
            values.append(v)

        self.__dict__['is_bound'] = False

    def __getattr__(self, name):
        """  eg:
             v = row.header
        """
        try:
            i = self._headers.get(name, name)
            return self.values[i]
        except (TypeError, IndexError) as e:
            raise AttributeError(self.__attr_err_msg(name)) from e

    def __getitem__(self, name):
        """ eg:
            v = row['header']

        name is converted to an integer, which is then used to reference self.values

        (if name is already an integer, it's faster to use row.values[i])
        (if name is a slice, it should be called directly on row.values)
        """
        try:
            i = self._headers.get(name, name)
            return self.values[i]
        except (TypeError, IndexError) as e:
            raise AttributeError(self.__attr_err_msg(name)) from e

    def __setattr__(self, name, value):
        """ eg:
            row.header = v
        """
        if name in self.__dict__:               # a property directly on flux_row_cls
            self.__dict__[name] = value
            return

        try:
            i = self._headers.get(name, name)
            self.values[i] = value
        except (TypeError, IndexError) as e:
            raise AttributeError(self.__attr_err_msg(name)) from e

    def __setitem__(self, name, value):
        """ eg:
            row['header'] = v

        name is converted to an integer, which is then used to reference self.values

        (if name is already an integer, it's faster to use row.values[i])
        (if name is a slice, it should be called directly on row.values)
        """
        try:
            i = self._headers.get(name, name)
            self.values[i] = value
        except (TypeError, IndexError) as e:
            raise AttributeError(self.__attr_err_msg(name)) from e

    def __getstate__(self):
        return self.__dict__

    def __setstate__(self, d):
        super().__setattr__('__dict__', d)

    def __attr_err_msg(self, name):
        if isinstance(name, slice):
            return 'slice should be used on row.values\n(eg, row.values[2:5], not row[2:5])'

        names = '\n\t'.join(self.names)
        return "No flux_row_cls column named '{}'\navailable columns:\n\t{}".format(name, names)

    def __len__(self):
        return len(self.values)

    def __iter__(self):
        return iter(self.values)

    def __repr__(self):
        i = self.__dict__.get('i')
        if i is not None:
            i = '({:,})  '.format(i)
        else:
            i = ''

        return 'flux_row  {}{}'.format(i, repr(self.values))


class lev_row_cls(flux_row_cls):

    class_names = {'_headers',
                   'is_bound',
                   'values',
                   'address'}

    def __init__(self, headers, values, address):
        super().__init__(headers, values)
        self.__dict__['address'] = address

    def __repr__(self):
        return self.address + ' ' + repr(self.values)




from . workbook import open_workbook
from . workbook import close_workbook
from . workbook import any_excel_instance
from . worksheet import get_worksheet

__all__ = ['open_workbook',
           'close_workbook',
           'any_excel_instance',
           'get_worksheet']




import copy

from math import ceil

from types import GeneratorType

from collections import OrderedDict
from collections import defaultdict
from collections import Iterable


def modify_iteration_depth(v, depth=0):
    """
    eg:
        'a' = modify_iteration_depth([['a']], depth=0)
        ['a', 'b', 'c'] = modify_iteration_depth([['a', 'b', 'c']], depth=0)

        [['a']] = modify_iteration_depth('a', depth=2)
        [['a', 'b', 'c']] = modify_iteration_depth(['a', 'b', 'c'], depth=2)
    """
    v  = generator_to_list(v)
    nd = iteration_depth(v)

    if nd > depth:
        # remove unneccesary nesting
        for _ in range(nd - depth):
            if is_subscriptable(v) and len(v) == 1:
                v = list(v)
                v = v[0]

    elif nd < depth:
        # add additional nesting
        for _ in range(depth - nd):
            v = [v]

    return v


def assert_iteration_depth(v, depth):
    """ make sure values have certain number of iterable layers """
    nd = iteration_depth(v)

    if nd != depth:
        msg = ('invalid iteration depth {}, value must have {} layers of iteration'
               .format(nd, depth))

        if depth == 1:
            msg += ', (ie, a list)'
        elif depth == 2:
            msg += ', (ie, list of lists)'

        raise IndexError(msg)


def iteration_depth(v):
    """ determine iteration-depth (number of dimensions)
    eg:
        0 = iteration_depth('a')
        1 = iteration_depth(['a'])
        1 = iteration_depth([])
        2 = iteration_depth(['a'], ['b']])
        2 = iteration_depth([[]])
    """
    if not is_iterable(v):
        return 0

    if isinstance(v, GeneratorType):
        v = list(v)

    if not is_subscriptable(v):
        return 1

    if len(v) == 0:
        return 1

    return 1 + iteration_depth(v[0])


def depth_one(v):
    """ add at least one iteration depth
    iteration of string characters dont count

    eg:
        for v in depth_one('abc'):
            v = 'abc'

        for v in depth_one(['abc']):
            b = 'abc'

        for v in depth_one({'a': 0, 'b': 1, 'c': 2}):
            v = (a, 0)
    """
    if isinstance(v, (list, tuple, GeneratorType)):
        return v

    if isinstance(v, dict):
        return v.items()

    return [v]


def is_iterable(v):
    """ determine if value can be iterated
    iteration of string characters dont count

    eg:
        False = is_iterable('mike')
        True  = is_iterable(['m', 'i' 'k' 'e'])
    """
    is_iter = isinstance(v, Iterable)
    is_str  = isinstance(v, str)

    return is_iter and not is_str


def is_subscriptable(v):
    return isinstance(v, (list, tuple))


def generator_to_list(v, recurse=False):
    if is_vengeance_class(v):
        v = list(v.rows())

    if isinstance(v, GeneratorType):
        v = list(v)

    if recurse:
        for _ in range(1, iteration_depth(v)):
            v = list(v)
            for i, _v_ in enumerate(v):
                v[i] = generator_to_list(_v_)

    return v


def stride_sequence(sequence, stride_len):
    sequence = generator_to_list(sequence)

    for i_1 in range(0, len(sequence), stride_len):
        i_2 = i_1 + stride_len
        yield sequence[i_1:i_2]


def divide_sequence(sequence, num_divisions):
    sequence = generator_to_list(sequence)

    num_items  = len(sequence)
    stride_len = ceil(num_items / num_divisions)
    stride_len = max(stride_len, 1)

    for i_1 in range(0, num_items, stride_len):
        i_2 = i_1 + stride_len
        yield sequence[i_1:i_2]


def is_vengeance_class(o):
    bases = {b.__name__ for b in o.__class__.mro()}
    if 'flux_cls' in bases or 'excel_levity_cls' in bases:
        return True

    return False


def is_flux_row_class(o):
    bases = {b.__name__ for b in o.__class__.mro()}
    if 'flux_row_cls' in bases or 'lev_row_cls' in bases:
        return True

    return False


def transpose(m):
    m = generator_to_list(m)

    if iteration_depth(m) == 1:
        num_r = 1
        num_c = len(m)
        t = [[m[r] for _ in range(num_r)]
                   for r in range(num_c)]
    else:
        num_r = len(m)
        num_c = len(m[0])
        t = [[m[c][r] for c in range(num_r)]
                      for r in range(num_c)]

    return t


def append_matrices(direction, *matrices, has_header=True):
    d = direction

    if d.startswith('v') or d.startswith('row'):
        return append_rows(*matrices, has_header=has_header)

    if d.startswith('h') or d.startswith('col'):
        return append_columns(*matrices)

    raise ValueError("invalid direction: '{}'".format(direction))


def append_rows(*matrices, has_header=True):

    def append(m_1, m_2):
        m_1 = generator_to_list(m_1)
        m_2 = generator_to_list(m_2)

        if len(m_1[0]) != len(m_2[0]):
            raise IndexError('vertical append requires matrices to have equal number of columns')

        if is_empty(m_1):
            return m_2

        if is_empty(m_2):
            return m_1

        m_a = copy.copy(m_1)
        if has_header:
            m_a.extend(m_2[1:])
        else:
            m_a.extend(m_2)

        return m_a

    m_f, matrices = matrices[0], matrices[1:]
    for m in matrices:
        m_f = append(m_f, m)

    return m_f


def append_columns(*matrices):

    def append(m_1, m_2):
        m_1 = generator_to_list(m_1)
        m_2 = generator_to_list(m_2)

        if len(m_1) != len(m_2):
            raise IndexError('horizontal append requires matrices to have equal number of rows')

        if is_empty(m_1):
            return m_2

        if is_empty(m_2):
            return m_1

        return [row_1 + row_2 for row_1, row_2 in zip(m_1, m_2)]

    m_f, matrices = matrices[0], matrices[1:]
    for m in matrices:
        m_f = append(m_f, m)

    return m_f


def is_empty(v):
    """ determine if iterable is composed entirely of empty iterables, eg, [] or [[]]
    eg:
        True  = is_empty( [] )
        True  = is_empty( [[]] )
        False = is_empty( [[None]] )
        False = is_empty( [[], [], []] )
    """
    m = generator_to_list(v)
    num_rows = len(m)
    num_cols = len(m[0])

    return (num_rows == 1) and (num_cols == 0)


def index_sequence(seq, start=0):
    items = []
    non_unique = defaultdict(int)

    for i, v in enumerate(seq, start):
        if v == '':
            v = 'None'

        if v in non_unique:
            v = '{} (nonunique_{})'.format(v, non_unique[v] + 1)

        items.append((v, i))
        non_unique[v] += 1

    return OrderedDict(items)


def ordered_unique(sequence):
    return list(OrderedDict((k, None) for k in sequence).keys())


def invert_mapping(d, inversion=None):
    """ return new mapping of
    {value: key}
    {key: value}

    eg:
        d = {'a': [-1, -1, -1]
             'b': 2
             'c': 3
             'd': 3}

        {-1: ['a', 'a', 'a']
         2: b
         3: ['c', 'd']} = invert_mapping(d)
    """
    if not isinstance(d, dict):
        raise TypeError('value must be a dictionary')

    inverted = OrderedDefaultDict(list)
    for k, v in d.items():
        _k_, _v_ = v, k

        if inversion:
            _k_ = inversion(_k_)

        for _k_ in depth_one(v):
            inverted[_k_].append(_v_)

    inverted = OrderedDict(inverted)
    for k, v in inverted.items():
        if len(v) == 1:
            inverted[k] = v[0]

    return inverted


class OrderedDefaultDict(OrderedDict):

    def __init__(self, default_factory=None, *a, **kw):
        if (default_factory is not None) and not callable(default_factory):
            raise TypeError('first argument must be callable')

        self.default_factory = default_factory

        if a and self.default_factory is list:
            self.__append_values(a[0])
        else:
            OrderedDict.__init__(self, *a, **kw)

    def __append_values(self, sequence):
        """
        special case when a sequence was submitted to constructor and
        all values should be appended as lists
        """

        if isinstance(sequence, dict):
            sequence = OrderedDict(sequence.items()).items()
        elif not isinstance(sequence, (list, tuple)):
            raise ValueError

        for item in sequence:
            if len(item) == 2:
                self[item[0]].append(item[1])
            elif len(item) < 2:
                raise ValueError('need more than 1 value to unpack')
            elif len(item) > 2:
                raise ValueError('too many values to unpack (expected 2)')

    def __getitem__(self, key):
        try:
            return OrderedDict.__getitem__(self, key)
        except KeyError:
            return self.__missing__(key)

    def __missing__(self, key):
        if self.default_factory is None:
            raise KeyError(key)

        self[key] = value = self.default_factory()

        return value

    def defaultdict(self):
        return defaultdict(self.default_factory, self.items())

    def ordereddict(self):
        return OrderedDict(self)

    def count_values(self):
        d = OrderedDict()
        for k, v in self.items():
            if is_iterable(v):
                d[k] = len(v)
            else:
                d[k] = 1

        return d

    def modify_iteration_depth(self, minimum=0):
        for k, v in self.items():
            self[k] = modify_iteration_depth(v, minimum)






# excel-specific constants
# (explicitly encoded as ascii instead of utf to avoid issues w/ C++ safearray strings)
xl_class_name   = 'XLMAIN'.encode('ascii')
xl_desk_class   = 'XLDESK'.encode('ascii')
xl_excel7_class = 'EXCEL7'.encode('ascii')

# HKEY_CLASSES_ROOT\Excel.Application: excel's clsid
xl_clsid = '{00020400-0000-0000-C000-000000000046}'

excel_errors = {-2146826281: 'error div0',
                -2146826246: 'error na',
                -2146826259: 'error name',
                -2146826288: 'error null',
                -2146826252: 'error num',
                -2146826265: 'error ref',
                -2146826273: 'error value',
                0x800A07FA:  'error div0'}

# colors
xl_yellow = 13434879
xl_clear  = -4142

# find parameters
xl_values     = -4163
xl_cell_type_formulas = -4123
xl_errors     = 16
xl_whole      = 1
xl_part       = 2
xl_next       = 1
xl_previous   = 2
xl_by_rows    = 1
xl_by_columns = 2
xl_up         = -4162
xl_to_right   = -4161

# enums
xl_paste_column_widths = 8
xl_range_ms_xml = 12
xl_range_value_ms_persist_xml = 12

# external / windows / shell constants
xl_maximized = -4137
sw_show      = 1
native_om    = -16

vb_ok_only = 0
process_terminate = 1




import json

from timeit import default_timer
from datetime import date
from collections import Iterable

nbsp = chr(160)
nonstandard_chars = "\"\'\n\t" + nbsp


def print_runtime(f):

    def runtime_wrapper(*args, **kwargs):
        f_name = f.__module__.split('.')[-1]
        f_name = '{}.{}'.format(f_name, f.__name__)

        tic = default_timer() * 1000

        if kwargs:
            result = f(*args, **kwargs)
        elif args:
            result = f(*args)
        else:
            result = f()

        toc = default_timer() * 1000
        elapsed = -(tic - toc)

        print_unicode('   τ: @{}: {}'.format(f_name, format_ms(elapsed)))

        return result

    return runtime_wrapper


def format_ms(ms):
    m, s = divmod(ms / 1000, 60)
    h, m = divmod(m, 60)

    if h >= 1:
        f_ms = '{} hr {} min'.format(h, m)
    elif m >= 1:
        f_ms = '{} min {} sec'.format(int(m), int(s))
    elif s > 1:
        f_ms = '{:.2f} sec'.format(s)
    elif ms > 10:
        f_ms = '{} ms'.format(int(ms))
    elif ms > 1:
        f_ms = '{:.1f} ms'.format(ms)
    else:
        f_ms = '{:.2f} us'.format(int(ms * 1000))

    return f_ms


def print_performance(f=None, *, repeat=3, number=100):
    """
    stolen and modified from
    https://github.com/realpython/materials/blob/master/pandas-fast-flexible-intuitive/tutorial/timer.py

    Decorator: prints time from best of repeat trials.

    Mimics timeit.repeat(), but avg. time is printed.
    Returns function result and prints time.

    You can decorate with or without parentheses, as in
    Python's @dataclass class decorator.
    """
    import gc
    import functools
    import itertools

    _repeat = functools.partial(itertools.repeat, None)

    def wrap(func):
        @functools.wraps(func)
        def performance_wrapper(*args, **kwargs):

            # Temporarily turn off garbage collection
            gcold = gc.isenabled()
            gc.disable()

            result = None
            best   = None
            total  = 0

            try:
                for _ in _repeat(repeat):
                    total = 0

                    for _ in _repeat(number):                   # number of trials within each repeat.
                        tic = default_timer() * 1000
                        result = func(*args, **kwargs)
                        toc = default_timer() * 1000
                        elapsed = -(tic - toc)

                        if best is None:
                            best = elapsed
                        else:
                            best = min(best, elapsed)

                        total += elapsed

                average = total / number

                f_name = func.__module__.split('.')[-1]
                f_name = '{}.{}'.format(f_name, func.__name__)

                print_unicode('   τ: @{}: {} (average), {} (best)'.format(f_name,
                                                                          format_ms(average),
                                                                          format_ms(best),))

            finally:
                if gcold:
                    gc.enable()

            return result

        return performance_wrapper

    if f is None:
        return wrap
    else:
        return wrap(f)


def repr_(sequence, concat=', ', quotes=False, wrap=True):
    """ extend formatting options beyond built-in repr() """

    def repr_recurse(v):
        if isinstance(v, str):
            if quotes:
                v = "'{}'".format(v)

            return v

        if is_sequence_dict:
            c = ': '
            w = False

            return repr_(v, c, quotes, w)

        if isinstance(v, Iterable):
            c = ', '
            w = wrap

            return repr_(v, c, quotes, w)

        return str(v)

    def wrapped_characters(v):
        if isinstance(v, dict):
            w = '{}'
        elif isinstance(v, tuple):
            w = '()'
        elif isinstance(v, list):
            w = '[]'
        else:
            w = None

        return w

    if wrap is True:
        w_ = wrapped_characters(sequence)
    else:
        w_ = wrap

    is_sequence_dict = isinstance(sequence, dict)
    if is_sequence_dict:
        sequence = sequence.items()

    s = concat.join(repr_recurse(o) for o in sequence)

    if w_:
        s = w_[0] + s + w_[1]

    return s


def sanitize(v):
    s = value_to_string(v)
    s = remove_multiple(s, nonstandard_chars)
    s = s.strip()

    return s


def value_to_string(v):
    if isinstance(v, str):
        return v

    if v is None:
        v = ''
    elif isinstance(v, date):
        v = v.strftime('%Y-%m-%d')
    else:
        v = str(v)

    return v


def remove_multiple(s, substrs, first_only=False):
    for sub in substrs:
        if first_only:
            s = s.replace(sub, '', 1)
        else:
            s = s.replace(sub, '')

    return s


def between(s, substr_1, substr_2=None):
    """ text between inner and outer substrings (exclusive)
    eg:
        'ell' = between('hello', 'h', 'o')
    """
    if substr_2 is None:
        substr_2 = substr_1

    i_1 = s.index(substr_1) + len(substr_1)
    i_2 = s.index(substr_2, i_1)

    return s[i_1:i_2]


def p_json_dumps(o, indent=4, ensure_ascii=False):
    """
    json can not convert certain python objects to string representations
        dates
        sets
    """
    def unhandled_conversion(_o):
        if isinstance(_o, date):
            return _o.isoformat()

        if isinstance(_o, set):
            return list(_o)

        raise TypeError('cannot convert type to json ' + repr(_o))

    s = json.dumps(o, indent=indent,  default=unhandled_conversion, ensure_ascii=ensure_ascii)

    return s


def to_ascii(s):
    greek_chrs = {'α': 'a',
                  'β': 'b',
                  'ε': 'e',
                  'τ': 't',
                  'ν': 'v'}

    s = str(s)
    for old, new in greek_chrs.items():
        s = s.replace(old, new)

    s = (s.encode('ascii', errors='backslashreplace')
          .decode('ascii'))

    return s


def print_unicode(s):
    try:
        print(s)
    except UnicodeEncodeError:
        print(to_ascii(s))


def vengeance_message(s):
    print_unicode('   ν: {}'.format(s))






import pyodbc
from .flux_cls import flux_cls


class odbc_service_cls:

    def __init__(self, cn_str, sql):
        self.cn_str = cn_str
        self.sql    = sql

        self._cn = pyodbc.connect(self.cn_str)
        self._cr = self._cn.cursor()

    def execute(self):
        self._cr.execute(self.sql)

    def rows(self):
        self.execute()

        m = [col[0] for col in self._cr.description]
        m.extend(list(row) for row in self._cr.fetchall())

        self.disconnect()

        return m

    def flux_rows(self):
        return flux_cls(self.rows())

    def disconnect(self):
        if self._cn is None:
            return

        if self._cr:
            self._cr.close()

        self._cn.close()
        self._cr = None
        self._cn = None


